@@ -33,6 +33,9 @@
 #include "library/gnome-keyring-private.h"
 #include "library/gnome-keyring-proto.h"
 
+#include "common/gkr-async.h"
+#include "common/gkr-location.h"
+
 #include <unistd.h>
 #include <fcntl.h>
 #include <errno.h>
@@ -66,17 +69,21 @@ typedef struct _GkrAskRequestPrivate GkrAskRequestPrivate;
 
 struct _GkrAskRequestPrivate {
 	GObject* object;
+	GQuark location;
 	
-	gchar* title;
-	gchar* primary;
-	gchar* secondary;
+	gchar *title;
+	gchar *primary;
+	gchar *secondary;
+	gchar *checktext;
+	gboolean location_selector;
 	
 	gboolean completed;
 	guint flags;
 	
 	gint ask_pid;
+	gint in_fd;
+	gint out_fd;
 	GkrBuffer buffer;
-	guint input_watch;
 };
 
 #define GKR_ASK_REQUEST_GET_PRIVATE(o)  \
@@ -236,7 +243,6 @@ mark_completed (GkrAskRequest *ask, GkrAskResponse resp)
 		if (resp)
 			ask->response = resp;
 		pv->completed = TRUE;
-		g_signal_emit (ask, signals[COMPLETED], 0);
 	}
 }
 
@@ -245,11 +251,6 @@ kill_ask_process (GkrAskRequest *ask)
 {
 	GkrAskRequestPrivate *pv = GKR_ASK_REQUEST_GET_PRIVATE (ask);
 	
-	if (pv->input_watch != 0) {
-
-		g_source_remove (pv->input_watch);
-		pv->input_watch = 0;
-	}
 	if (pv->ask_pid != 0) {
 		kill (pv->ask_pid, SIGKILL);
 		pv->ask_pid = 0;
@@ -270,15 +271,16 @@ cancel_ask_if_active (GkrAskRequest *ask)
 }
 
 static void
-finish_ask_io (GkrAskRequest *ask, gboolean failed)
+finish_ask_io (GkrAskRequest *ask, gboolean success)
 {
 	GkrAskRequestPrivate *pv;
-	gchar *line, *next;
+	gchar *line, *next, *value;
+	GError *error = NULL;
+	GKeyFile *key_file;
 	int i;
 
 	pv = GKR_ASK_REQUEST_GET_PRIVATE (ask);
 	
-	pv->input_watch = 0;
 	pv->ask_pid = 0;
 
 	/* Cleanup for response processing */
@@ -288,7 +290,7 @@ finish_ask_io (GkrAskRequest *ask, gboolean failed)
 	ask->original_password = NULL;
 	
 	/* A failed request */
-	if (failed) {
+	if (!success) {
 		mark_completed (ask, GKR_ASK_RESPONSE_FAILURE);
 		return;
 	}
@@ -298,34 +300,51 @@ finish_ask_io (GkrAskRequest *ask, gboolean failed)
 	 * to a minimum, for security reasons.
 	 */
 	gkr_buffer_add_byte (&pv->buffer, 0);
-	for (i = 0, line = (gchar*)pv->buffer.buf; i < 4; ++i) {
+	for (i = 0, line = (gchar*)pv->buffer.buf; line && i < 2; ++i) {
 		
 		/* Break out the line */
 		next = strchr (line, '\n');
-		if (next) 
+		if (next) {
 			*next = 0;
-		
-		/* First line is the response */
-		if (i == 0) {
-			ask->response = atol (line);
-			if (ask->response < GKR_ASK_RESPONSE_ALLOW)
-				break;
-				
-		/* Next line is the typed password (if any) */
-		} else if (i == 1) {
-			ask->typed_password = gnome_keyring_memory_strdup (line);
-			
-		/* Last line is the original password (if any) */
-		} else if (i == 2) {
-			ask->original_password = gnome_keyring_memory_strdup (line);
+			++next;
 		}
 		
-		/* No more lines? */
-		if (!next)
-			break;
-		line = next + 1;
+		/* First line is the password */
+		if (i == 0)
+			ask->typed_password = gnome_keyring_memory_strdup (line);
+		
+		/* Second line is the original password (if any)*/
+		else if (i == 1)
+			ask->original_password = gnome_keyring_memory_strdup (line);
+			
+		line = next;
 	}
 	
+	if (!line) {
+		g_warning ("missing dialog response from ask tool");
+		mark_completed (ask, GKR_ASK_RESPONSE_FAILURE);
+		return;
+	}
+	
+	/* The remainder is a GKeyFile */
+	key_file = g_key_file_new ();
+	if (!g_key_file_load_from_data (key_file, line, strlen (line), G_KEY_FILE_NONE, &error)) {
+		g_warning ("couldn't parse dialog response from ask tool: %s", error->message);
+		g_error_free (error);
+		g_key_file_free (key_file);
+		mark_completed (ask, GKR_ASK_RESPONSE_FAILURE);
+		return;
+	}
+	
+	/* Pull out some values */
+	ask->checked = g_key_file_get_boolean (key_file, "check", "check-active", NULL);
+	ask->response = g_key_file_get_integer (key_file, "general", "response", NULL);
+	value = g_key_file_get_value (key_file, "location", "location-selected", NULL);
+	if (value)
+		ask->location_selected = gkr_location_from_string (value);
+	g_free (value);
+	g_key_file_free (key_file);
+		
 	/* An invalid result from the ask tool */
 	if (!ask->response) {
 		mark_completed (ask, GKR_ASK_RESPONSE_FAILURE);
@@ -346,46 +365,153 @@ finish_ask_io (GkrAskRequest *ask, gboolean failed)
 	g_object_unref (ask);
 }
 
-static gboolean
-ask_io (GIOChannel *channel, GIOCondition cond, gpointer data)
+static void
+close_fd (gpointer data)
 {
-	GkrAskRequest *ask;
-	GkrAskRequestPrivate *pv;
-	gboolean ret = TRUE;
-	guchar *buffer;
-	int res;
-	int fd;
+	int *fd = (int*)data;
+	g_assert (fd);
+	close (*fd);
+	*fd = -1;
+}
 
-	ask = GKR_ASK_REQUEST (data);
-	pv = GKR_ASK_REQUEST_GET_PRIVATE (ask);
+static gchar*
+prep_dialog_data (GkrAskRequest *ask)
+{
+	GkrAskRequestPrivate *pv = GKR_ASK_REQUEST_GET_PRIVATE (ask);
+	GkrLocationManager *locmgr;
+	gboolean need_locations = FALSE;
+	GKeyFile *key_file;
+	gchar *value, *data;
+	GArray *array;
+	GSList *locations, *l;
+	GParamSpec *spec;
+	GQuark loc;
+	const gchar *t;
+	
+	key_file = g_key_file_new ();
+	
+	value = format_object_markup (pv->object, pv->title, NULL);
+	g_key_file_set_value (key_file, "general", "title", value);
+	g_free (value);
+	
+	value = format_object_markup (pv->object, pv->primary, NULL);
+	g_key_file_set_value (key_file, "general", "primary", value);
+	g_free (value);
+	
+	value = format_object_markup (pv->object, pv->secondary, NULL);
+	g_key_file_set_value (key_file, "general", "secondary", value);
+	g_free (value);
+	
+	g_key_file_set_integer (key_file, "general", "flags", pv->flags);
+
+	if (pv->checktext) {
+		g_key_file_set_boolean (key_file, "check", "check-enable", TRUE);
+		g_key_file_set_value (key_file, "check", "check-text", pv->checktext);
+	}
+	
+	/* Display the location drop down selector */	
+	if (pv->location_selector) {
+		g_key_file_set_boolean (key_file, "location", "location-selector", TRUE);
+		need_locations = TRUE;
+	}
+
+	if (!pv->location && pv->object) {
+		spec = g_object_class_find_property (G_OBJECT_GET_CLASS (pv->object), "location");
+		if (spec)
+			g_object_get (pv->object, "location", &pv->location, NULL);
+	}
+
+	/* See if we should send a location to display */
+	if (pv->location) {
+		loc = gkr_location_get_base (pv->location);
+		
+		/* Suppress local stuff unless displying the selector */ 
+		if (!need_locations && loc == GKR_LOCATION_BASE_LOCAL)
+			loc = 0;
+			
+		if (loc) {
+			g_key_file_set_value (key_file, "location", "location", 
+			                      gkr_location_to_string (loc));
+			need_locations = TRUE;
+		}
+	}
+	
+	if (need_locations) {
+		locmgr = gkr_location_manager_get ();
+		array = g_array_new (TRUE, TRUE, sizeof (gchar*));
+		locations = gkr_location_manager_get_base_locations (locmgr);
+			
+		/* Send all the locations and all the display names */
+		for (l = locations; l; l = g_slist_next (l)) {
+		 	t = gkr_location_to_string (GPOINTER_TO_UINT (l->data)); 
+			g_array_append_val (array, t);
+		}
+		g_key_file_set_string_list (key_file, "location", "names", (const gchar**)array->data, array->len);
+		
+		g_array_set_size (array, 0);
+		for (l = locations; l; l = g_slist_next (l)) {
+			t = gkr_location_manager_get_base_display (locmgr, GPOINTER_TO_UINT (l->data)); 
+			g_array_append_val (array, t);
+		}
+		g_key_file_set_string_list (key_file, "location", "display-names", (const gchar**)array->data, array->len);
+
+		g_array_free (array, TRUE);
+		g_slist_free (locations);
+	}
+
+	data = g_key_file_to_data (key_file, NULL, NULL);
+	g_return_val_if_fail (data, NULL);
+	
+	return data;
+}
+
+static gboolean
+read_until_end (GkrAskRequest *ask)
+{
+	GkrAskRequestPrivate *pv = GKR_ASK_REQUEST_GET_PRIVATE (ask);
+	gboolean ret = FALSE;
+	guchar *buf;
+	int res;
+	
+	g_return_val_if_fail (pv->out_fd >= 0, FALSE);
 	
 	/* Passwords come through this buffer */
-	buffer = gnome_keyring_memory_alloc (128);
+	buf = gnome_keyring_memory_alloc (128);
 
-	if (ret && (cond & G_IO_IN)) {
-		do 
-		{
-			fd = g_io_channel_unix_get_fd (channel);
-			res = read (fd, buffer, 128);
-			if (res < 0) {
-				if (errno != EINTR && errno != EAGAIN) {
-					finish_ask_io (ask, TRUE);
-					ret = FALSE;
-					break;
-				}
-			} else if (res > 0) {
-				gkr_buffer_append (&pv->buffer, buffer, res);
-			}
-		} while (res > 0);
-	}
+	gkr_async_register_cancel (close_fd, &pv->out_fd);
 
-	if (ret && (cond & G_IO_HUP)) {	
-		finish_ask_io (ask, FALSE);
-		ret = FALSE;
+	for (;;) {
+
+		if (gkr_async_is_stopping ())
+			break;
+
+		gkr_async_begin_concurrent ();
+			res = read (pv->out_fd, buf, 128);
+		gkr_async_end_concurrent ();
+
+		/* Got an error */
+		if (res < 0) {
+			if (errno == EINTR && errno == EAGAIN) 
+				continue;
+			g_warning ("couldn't read from ask tool: %s", g_strerror (errno));
+			break;
+			
+		/* Got some data */
+		} else if (res > 0) {
+			gkr_buffer_append (&pv->buffer, buf, res);
+			
+		/* End of data */
+		} else if (res == 0) {
+			ret = TRUE;
+			break;
+		}
 	}
 	
-	gnome_keyring_memory_free (buffer);
+	gnome_keyring_memory_free (buf);
+	gkr_async_unregister_cancel (close_fd, &pv->out_fd);
 	
+	close_fd (&pv->out_fd);
+
 	return ret;
 }
 
@@ -393,17 +519,14 @@ static gboolean
 launch_ask_helper (GkrAskRequest *ask)
 {
 	GkrAskRequestPrivate *pv = GKR_ASK_REQUEST_GET_PRIVATE (ask);
-	GIOChannel *channel;
 	const gchar* display;
 	char **envp;
 	int i, n;
-	int stdout_fd;
 	GError *error = NULL;
 	char *argv[] = {
 		LIBEXECDIR "/gnome-keyring-ask",
 		NULL,
 	};
-	gboolean res;
 
 	/* Calculate us some environment */
 	i = 0;
@@ -412,7 +535,7 @@ launch_ask_helper (GkrAskRequest *ask)
 	n = i;
 	
 	/* Any environment we have */
-	envp = g_new (char*, n + 1 + 6);
+	envp = g_new (char*, n + 2);
 	for (i = 0; i < n; i++)
 		envp[i] = g_strdup (environ[i]);
 	
@@ -420,28 +543,69 @@ launch_ask_helper (GkrAskRequest *ask)
 	display = gkr_ask_daemon_get_display ();
 	if (display && display[0])
 		envp[i++] = g_strdup_printf ("DISPLAY=%s", display);
-	envp[i++] = format_object_markup (pv->object, "ASK_TITLE=", pv->title);
-	envp[i++] = format_object_markup (pv->object, "ASK_PRIMARY=", pv->primary);
-	envp[i++] = format_object_markup (pv->object, "ASK_SECONDARY=", pv->secondary);
-	envp[i++] = g_strdup_printf ("ASK_FLAGS=%d", pv->flags);
 	envp[i++] = NULL;
 
 	gkr_buffer_resize (&pv->buffer, 0);
 	
 	if (!g_spawn_async_with_pipes (NULL, argv, envp, 0, NULL, NULL, &pv->ask_pid, 
-	                               NULL, &stdout_fd, NULL, &error)) {
+	                               &pv->in_fd, &pv->out_fd, NULL, &error)) {
 		g_warning ("couldn't spawn gnome-keyring-ask tool: %s", 
 		           error && error->message ? error->message : "unknown error");
-		res = FALSE;
-	} else {
-		channel = g_io_channel_unix_new (stdout_fd);
-		pv->input_watch = g_io_add_watch (channel, G_IO_IN | G_IO_HUP, ask_io, ask);
-		g_io_channel_unref (channel);
-		res = TRUE;
-	}
+		pv->out_fd = -1;
+		pv->in_fd = -1;
+		return FALSE;
+	} 
 	
 	g_strfreev (envp);
-	return res;
+	return TRUE;
+}
+
+static gboolean
+send_all_data (GkrAskRequest *ask, const gchar *buf, gsize len)
+{
+	GkrAskRequestPrivate *pv = GKR_ASK_REQUEST_GET_PRIVATE (ask);
+	gboolean ret = FALSE;
+	int res;
+	
+	g_return_val_if_fail (pv->in_fd >= 0, FALSE);
+	
+	gkr_async_register_cancel (close_fd, &pv->in_fd);
+
+	while (len > 0) {
+
+		if (gkr_async_is_stopping ())
+			break;
+
+		gkr_async_begin_concurrent ();
+			res = write (pv->in_fd, buf, len);
+		gkr_async_end_concurrent ();
+
+		/* Got an error */
+		if (res < 0) {
+			if (errno == EINTR && errno == EAGAIN) 
+				continue;
+			g_warning ("couldn't write data to ask tool: %s", g_strerror (errno));
+			break;
+			
+		/* Got some data */
+		} else if (res > 0) {
+			len -= res;
+			buf += res;
+			
+		/* Eh? */
+		} else if (res == 0) {
+			g_warning ("couldn't write data to ask tool");
+			break;
+		}
+	}
+	
+	if (len == 0)
+		ret = TRUE;
+	
+	gkr_async_unregister_cancel (close_fd, &pv->in_fd);
+	
+	close_fd (&pv->in_fd);
+	return ret;
 }
 
 static void 
@@ -489,6 +653,10 @@ gkr_ask_request_init (GkrAskRequest *ask)
 	pv->title = g_strdup ("");
 	pv->primary = g_strdup ("");
 	pv->secondary = g_strdup ("");
+	pv->checktext = NULL;
+	
+	pv->out_fd = -1;
+	pv->in_fd = -1;
 	
 	/* Use a secure memory buffer */
 	gkr_buffer_init_full (&pv->buffer, 128, gnome_keyring_memory_realloc);
@@ -515,6 +683,11 @@ gkr_ask_request_dispose (GObject *obj)
 	gnome_keyring_free_password (ask->typed_password);
 	ask->typed_password = NULL;
 	
+	if (pv->in_fd >= 0)
+		close_fd (&pv->in_fd);
+	if (pv->out_fd >= 0)
+		close_fd (&pv->out_fd);
+	
 	G_OBJECT_CLASS(gkr_ask_request_parent_class)->dispose (obj);
 }
 
@@ -530,15 +703,14 @@ gkr_ask_request_finalize (GObject *obj)
 	g_free (pv->title);
 	g_free (pv->primary);
 	g_free (pv->secondary);
-	pv->title = pv->primary = pv->secondary = NULL;
+	g_free (pv->checktext);
+	pv->title = pv->primary = pv->secondary = pv->checktext = NULL;
 	
 	g_assert (pv->ask_pid == 0);
+	g_assert (pv->in_fd < 0);
+	g_assert (pv->out_fd < 0);
 	
 	gkr_buffer_uninit (&pv->buffer);
-	
-	if (pv->input_watch) 
-		g_source_remove (pv->input_watch);
-	pv->input_watch = 0;
 
 	G_OBJECT_CLASS(gkr_ask_request_parent_class)->finalize (obj);
 }
@@ -561,11 +733,6 @@ gkr_ask_request_class_init (GkrAskRequestClass *klass)
 			G_SIGNAL_RUN_LAST, G_STRUCT_OFFSET (GkrAskRequestClass, check_request),
 			accumulate_checks, NULL, gkr_ask_marshal_UINT__VOID, 
 			G_TYPE_UINT, 0);
-
-	signals[COMPLETED] = g_signal_new ("completed", GKR_TYPE_ASK_REQUEST, 
-			G_SIGNAL_RUN_FIRST, G_STRUCT_OFFSET (GkrAskRequestClass, completed),
-			NULL, NULL, g_cclosure_marshal_VOID__VOID, 
-			G_TYPE_NONE, 0);
 }
 
 /* -----------------------------------------------------------------------------
@@ -604,6 +771,17 @@ gkr_ask_request_set_secondary (GkrAskRequest *ask, const gchar *secondary)
 	pv->secondary = g_strdup (secondary);
 }
 
+void
+gkr_ask_request_set_check_option (GkrAskRequest *ask, const gchar *check_text)
+{
+	GkrAskRequestPrivate *pv = GKR_ASK_REQUEST_GET_PRIVATE (ask);
+	
+	g_assert (GKR_IS_ASK_REQUEST (ask));
+	
+	g_free (pv->checktext);
+	pv->checktext = g_strdup (check_text);
+}
+
 GObject*
 gkr_ask_request_get_object (GkrAskRequest *ask)
 {
@@ -673,6 +851,8 @@ void
 gkr_ask_request_prompt (GkrAskRequest *ask)
 {
 	GkrAskRequestPrivate *pv = GKR_ASK_REQUEST_GET_PRIVATE (ask);
+	gboolean ret;
+	gchar *data;
 	
 	g_assert (GKR_IS_ASK_REQUEST (ask));
 	
@@ -680,15 +860,27 @@ gkr_ask_request_prompt (GkrAskRequest *ask)
 	if (pv->completed) 
 		return;
 	
-	launch_ask_helper (ask);
+	ret = launch_ask_helper (ask);
+	if (ret) {
+		data = prep_dialog_data (ask);
+		g_return_if_fail (data);
+		ret = send_all_data (ask, data, strlen (data));
+		g_free (data);
+	}
+	if (ret)
+		ret = read_until_end (ask);
+	finish_ask_io (ask, ret);
 }
 
 void
 gkr_ask_request_cancel (GkrAskRequest *ask)
 {
+	GkrAskRequestPrivate *pv = GKR_ASK_REQUEST_GET_PRIVATE (ask);
 	g_assert (GKR_IS_ASK_REQUEST (ask));
 	
 	cancel_ask_if_active (ask);
+	if (!pv->completed)
+		mark_completed (ask, GKR_ASK_RESPONSE_FAILURE);
 }
 
 gboolean
@@ -698,3 +890,19 @@ gkr_ask_request_is_complete (GkrAskRequest *ask)
 	g_assert (GKR_IS_ASK_REQUEST (ask));
 	return pv->completed;
 }
+
+void
+gkr_ask_request_set_location_selector (GkrAskRequest *ask, gboolean have)
+{
+	GkrAskRequestPrivate *pv = GKR_ASK_REQUEST_GET_PRIVATE (ask);
+	g_assert (GKR_IS_ASK_REQUEST (ask));
+	pv->location_selector = have;
+}
+                                                          
+void
+gkr_ask_requset_set_location (GkrAskRequest *ask, GQuark loc)
+{
+	GkrAskRequestPrivate *pv = GKR_ASK_REQUEST_GET_PRIVATE (ask);
+	g_assert (GKR_IS_ASK_REQUEST (ask));
+	pv->location = loc;
+}
